import ExpressHandlebars from "./express-handlebars";
import type { ConfigOptions, Engine } from "../types";
export { ExpressHandlebars };
export declare function create(config?: ConfigOptions): ExpressHandlebars;
export declare function engine(config?: ConfigOptions): Engine;
